#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Contains the minimal necessary set of routines to run Ogma.
"""

# ---------------------------------------------------------------------------
# Standard imports:
import sys
import argparse
import os
import importlib
import importlib.util

# Third party imports
try:
    import colored_traceback

    colored_traceback.add_hook(style="native")
except ImportError:
    pass

import colorama
import argcomplete


# Local imports
from . import modelutils
from . import utils
from . import commands
from . import version

# ---------------------------------------------------------------------------

colorama.init()

__description__ = "Ogma: A Database Access Code Generator for Java"

# An attempt to find a default output directory.
DEFAULT_OUTPUT_DIR = os.path.abspath(os.path.join(".", "output"))

# Default base package for generated code
DEFAULT_BASE_PACKAGE = "com.example.dbutils"


def completion_code(args):
    print(argcomplete.shellcode([args.program], args.shell))


def add_modelutils(model):
    if hasattr(modelutils, "__all__"):
        all_names = modelutils.__all__
    else:
        all_names = [name for name in dir(modelutils) if not name.startswith("_")]

    globs = vars(model)
    globs.update({name: getattr(modelutils, name) for name in all_names})


def load_module(file_path, name, decorator=None, delete_cache=False):
    spec = importlib.util.spec_from_file_location(name, file_path)
    module = importlib.util.module_from_spec(spec)
    if decorator is not None:
        decorator(module)
    spec.loader.exec_module(module)
    if delete_cache:
        # Delete the cache for future invocations
        cache = getattr(module, "__cached__", None)
        if cache and os.path.exists(cache):
            os.remove(cache)
    return module


def load_dbmodel(file_path, allow_imports):
    """
    Load a give python file as the dbmodel module
    """

    utils.print_action("Loading database model")
    try:
        modelutils.validation.validate_model_file(file_path, allow_imports)
        utils.li(os.path.abspath(file_path))
        # Make sure the module is always loaded
        importlib.invalidate_caches()
        dbmodel = load_module(
            file_path, "dbmodel", decorator=add_modelutils, delete_cache=True
        )
        return dbmodel
    except modelutils.validation.DatabaseModelException as dbmex:
        utils.error(str(dbmex))
    except Exception as ex:
        utils.print_end_action(str(ex))
        raise
    else:
        utils.print_end_action()


def parse_args(argv):
    """
    Parse and validate command line arguments
    """

    parser = argparse.ArgumentParser(description=__description__)

    parser.add_argument(
        "--version", action="version", version="%(prog)s {}".format(version.VERSION)
    )
    subparsers = parser.add_subparsers(title="Subcommands", dest="command_name")

    # A parent parser for common database access parameters
    # ------------------------------------------------------------------------
    dbopts_parser = argparse.ArgumentParser(add_help=False)
    dbopts_group = dbopts_parser.add_argument_group(title="Database access parameters")
    dbopts_group.add_argument(
        "--db-user",
        "-u",
        required=True,
        help="The database user to connect to the database",
    )
    dbopts_group.add_argument(
        "--db-password",
        "-p",
        required=True,
        help="The database password to connect to the database",
    )
    dbopts_group.add_argument(
        "--db-host", "-H", default="localhost", help="The host holding the database"
    )
    dbopts_group.add_argument(
        "--db-name", help="The name of the database to create or read"
    )
    dbopts_group.add_argument(
        "--db-port", "-P", type=int, default=3306, help="The database port"
    )

    # A parent parser for subcommands that take a model file
    # ------------------------------------------------------------------------
    model_parser = argparse.ArgumentParser(add_help=False)
    # Hidden argument to override the schema from the model:
    model_parser.add_argument("--schema", help=argparse.SUPPRESS)
    model_parser.add_argument(
        "--allow-imports",
        action="store_true",
        help="Allow import statements in the model file",
    )
    model_parser.add_argument(
        "model_file", metavar="MODEL_FILE", help="The file with the DB model"
    )

    # Subcommand: generate
    # ------------------------------------------------------------------------
    gen_parser = subparsers.add_parser(
        "generate", parents=[dbopts_parser, model_parser], help="Generate database code"
    )

    outdir_parser = gen_parser.add_argument_group("Output directories")

    # Output directories
    outdir_parser.add_argument(
        "--code-dir",
        "-c",
        default=DEFAULT_OUTPUT_DIR,
        help="The directory under which generated code should be",
    )
    outdir_parser.add_argument(
        "--sql-dir",
        default=DEFAULT_OUTPUT_DIR,
        help="The directory under which generated SQL should be",
    )
    outdir_parser.add_argument(
        "--config-dir",
        "-x",
        default=DEFAULT_OUTPUT_DIR,
        help="The directory under which generated config should be",
    )

    # Java / JOOQ specific
    gen_parser.add_argument(
        "--java", default=None, help="The path to the java binary that will run jOOQ"
    )
    gen_parser.add_argument(
        "--java-package",
        default=DEFAULT_BASE_PACKAGE,
        help="The base Java package of generated database code",
    )
    gen_parser.add_argument(
        "--classpath", help="The classpath, where to find jOOQ and the DB connector"
    )
    gen_parser.add_argument(
        "--no-jooq",
        default=False,
        action="store_true",
        help="Skip jooq code generation",
    )

    gen_parser.add_argument(
        "--keep-db",
        default=False,
        action="store_true",
        help="Do not delete the temporary database afterwards",
    )
    gen_parser.set_defaults(function=commands.generate)

    # Subcommand: enum-tables
    # ------------------------------------------------------------------------
    etab = subparsers.add_parser(
        "enum-tables",
        parents=[dbopts_parser, model_parser],
        help="Create tables with enum names",
    )
    etab.set_defaults(function=commands.enum_tables)
    etab.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Show more information about what is happening",
    )

    # Subcommand: enum-usage
    # ------------------------------------------------------------------------
    eusg = subparsers.add_parser(
        "enum-usage", parents=[model_parser], help="Report enum usage"
    )
    eusg.set_defaults(function=commands.enum_usage)

    # Subcommand: get-db-name
    # ------------------------------------------------------------------------
    gdbn = subparsers.add_parser(
        "get-db-name", help="Get a unique database name (for temp db)"
    )
    gdbn.set_defaults(function=commands.get_db_name)

    # Subcommand: create-db
    # ------------------------------------------------------------------------
    cdb = subparsers.add_parser(
        "create-db",
        parents=[dbopts_parser, model_parser],
        help="Create the database from the model",
    )
    cdb.set_defaults(function=commands.create_db)

    # Subcommand: drop-db
    # ------------------------------------------------------------------------
    ddb = subparsers.add_parser(
        "drop-db", parents=[dbopts_parser], help="Drop a given database"
    )
    ddb.set_defaults(function=commands.drop_db)
    ddb.add_argument("database")

    # Subcommand: completion
    # ------------------------------------------------------------------------
    compl = subparsers.add_parser("completion", help="Provide shell completion code")
    compl.set_defaults(function=completion_code)
    compl.add_argument(
        "shell", choices=["bash", "fish", "tcsh"], help="Name of the shell"
    )

    # ------------------------------------------------------------------------
    argcomplete.autocomplete(parser)
    args = parser.parse_args(argv[1:])
    if not hasattr(args, "function"):
        parser.error("Subcommand is missing")

    # If DB access is required, build a DbSettings object with the connection
    # parameters.
    if hasattr(args, "db_host"):
        args.dbsettings = modelutils.DbSettings(
            host=args.db_host,
            name=(
                args.db_name
                if args.db_name is not None
                else modelutils.get_new_database_name()
            ),
            user=args.db_user,
            password=args.db_password,
            port=args.db_port,
        )

    # Booleans starting with "no" end here!
    if hasattr(args, "no_jooq"):
        args.do_jooq = not args.no_jooq
        del args.no_jooq

    # If a model file was provided, load it as a python module
    if hasattr(args, "model_file"):
        args.dbmodel = load_dbmodel(args.model_file, args.allow_imports)

        # Override the schema if the hidden --schema option was provided.
        # WARNING: This hides problems in invalid models
        if hasattr(args, "schema") and args.schema is not None:
            args.dbmodel.schema_name.value = args.schema

    # Load the classpath from the environment if one was not provided in the
    # command line
    if not hasattr(args, "classpath") and "CLASSPATH" in os.environ:
        args.classpath = os.environ["CLASSPATH"]

    args.program = parser.prog
    return args


def main(argv=None):
    """ Run this program """
    if argv is None:
        argv = sys.argv
    args = parse_args(argv)
    try:
        args.function(args)
    except KeyboardInterrupt:
        sys.exit(-1)


if __name__ == "__main__":
    sys.exit(main(sys.argv) or 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ---------------------------------------------------------------------------
# Standard imports:
import collections
from enum import Enum

# Third party imports
import sqlalchemy

# Local imports
from .. import modelutils

# ---------------------------------------------------------------------------


def get_db_connection_string(dbsettings):
    db = dbsettings
    return f"mysql+pymysql://{db.user}:{db.password}@{db.host}:{db.port}/{db.name}"


def get_db_engine(dbsettings):
    return sqlalchemy.create_engine(get_db_connection_string(dbsettings))


class TypeFamilies(Enum):
    enum = 1
    boolean = 2
    binary = 3


def get_type_mappings(metadata, filter_types=None):
    """
    Return a double nested mapping of:
        table -> column -> type
    """
    if filter_types is None:
        filter_types = [TypeFamilies.enum, TypeFamilies.boolean]
    column_types = collections.defaultdict(dict)

    def visitor(column_name, column):
        # ENUMS
        if TypeFamilies.enum in filter_types and isinstance(
            column.type, modelutils.IntegerEnum
        ):
            column_types[column.table.name][column.name] = column.type.field_data.name
            return
        # BOOLEANS
        if TypeFamilies.boolean in filter_types and isinstance(
            column.type, modelutils.Boolean
        ):
            column_types[column.table.name][column.name] = "BOOLEAN"
            return
        # BINARIES
        if TypeFamilies.binary in filter_types and isinstance(
            column.type, modelutils.BINARY
        ):
            column_types[column.table.name][column.name] = "BINARY"

    metadata.visit_columns(visitor)

    def sort_by_keys(d):
        """
        Return an OrderedDict with the elements inserted according to the
        sorted keys.
        """
        return collections.OrderedDict(((k, d[k]) for k in sorted(d.keys())))

    # Sort by table name
    ordered_types = sort_by_keys(column_types)

    # And each table by column name
    for name in ordered_types:
        ordered_types[name] = sort_by_keys(column_types[name])

    return ordered_types



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import colored
import sys


def colstr(color, string):
    return f"{color}{string}{colored.style.RESET}"


def li(item, bullet=" *", color=colored.fore.LIGHT_GREEN, printer=print):
    """Print a list item"""
    printer(colstr(color, bullet), item)


def print_action(action, printer=print):
    li(action, bullet="===>", printer=printer)


def print_end_action(error=None, printer=print):
    color = colored.fore.DARK_GREEN if error is None else colored.fore.LIGHT_RED
    if error is None:
        error = ""
    li(error, color=color, bullet="===<", printer=printer)


def print_generated_file(file_path, printer=print):
    li(file_path, bullet="@", color=colored.fore.LIGHT_YELLOW, printer=printer)


def print_section_header(name, printer=print):
    line = colstr(colored.fore.LIGHT_BLUE, "=" * 80)
    printer()
    printer(line)
    printer("    " + name)
    printer(line)


def error(message, rc=1, printer=print):
    printer(colstr(colored.fore.LIGHT_RED, "ERROR:"), message)
    sys.exit(rc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Handling of the enum-usage subcommand
"""

# ---------------------------------------------------------------------------
# Standard imports:
import collections

# Third party imports

# Local imports
from .. import utils
from . import common

# ---------------------------------------------------------------------------


def enum_usage(args):
    eusage = common.get_type_mappings(args.dbmodel.metadata, [common.TypeFamilies.enum])
    enums = collections.defaultdict(list)
    for table, fields in eusage.items():
        for column, enum in fields.items():
            enums[enum].append((table, column))
    for enum, usages in enums.items():
        print(f"Enum: {enum}")
        for table, column in usages:
            utils.li(f"{table}.{column}")



#!/usr/bin/env python
# encoding: utf-8
"""
Handling of the generate subcommand
"""
# ---------------------------------------------------------------------------
# Standard imports:
import os
import subprocess
import shutil

# Local imports
from .. import codegen
from .. import modelutils
from .. import utils
from . import common


# ---------------------------------------------------------------------------

ENUM_PACKAGE_TPL = "{base}.{schema}.enums"
ENUM_CONVERTER_PACKAGE_TPL = ENUM_PACKAGE_TPL + ".converters"
DB_QUERY_PACKAGE_TPL = "{base}.{schema}.db"


def _generate_db_schema_ddl(metadata, schema, output_dir):
    utils.print_section_header("Database Creation DDL")
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    engines = ("mysql",)
    for engine in engines:
        file_name = f"full_ddl.{schema.lower()}.{engine}.sql"
        file_path = os.path.join(output_dir, file_name)
        with open(file_path, "w") as fsqlout:
            fsqlout.write(metadata.ddl(engine))
        utils.print_action(f"Generating database schema for {engine}")
        utils.print_generated_file(file_name)
        utils.print_end_action()


def _find_java():
    """
    Provide the java executable path, searching first in JAVA_HOME, then in the
    PATH
    """
    java_home = os.environ.get("JAVA_HOME", None)
    if java_home is not None:
        java_path = os.path.join(java_home, "bin", "java")
        if os.path.isfile(java_path) and os.access(java_path, os.X_OK):
            return java_path
    java_path = shutil.which("java")
    return java_path


def _run_jooq(java, config_file, classpath=None):
    """
    Run the jOOQ generator with the give configuration file and classpath
    """
    if java is None:
        # No java path was explicitly provided, so search
        java = _find_java()
    command = [java]
    if classpath is not None:
        command.append("-classpath")
        command.append(classpath)
    command.append("org.jooq.util.GenerationTool")
    command.append(config_file)
    print(" ".join(command))
    try:
        subprocess.check_call(command)
    except subprocess.CalledProcessError as cpe:
        return cpe


def generate(args):
    modelutils.validate_model(args.dbmodel)

    schema_name = args.dbmodel.schema_name()
    package_schema_name = schema_name.lower()

    # Add schema to packages
    enum_package = ENUM_PACKAGE_TPL.format(
        base=args.java_package, schema=package_schema_name
    )
    enum_conv_package = ENUM_CONVERTER_PACKAGE_TPL.format(
        base=args.java_package, schema=package_schema_name
    )
    db_query_package = DB_QUERY_PACKAGE_TPL.format(
        base=args.java_package, schema=package_schema_name
    )

    code_generator = codegen.EnumCodeGenerator(
        enums=args.dbmodel.enums,
        code_dir=args.code_dir,
        config_dir=args.config_dir,
        enum_package=enum_package,
        converter_package=enum_conv_package,
        model_file=args.model_file,
    )

    # Generate Java enums and Converters for jOOQ
    code_generator.generate_enum_java_code()

    # Generate DB scripts for all used engines
    _generate_db_schema_ddl(args.dbmodel.metadata, package_schema_name, args.sql_dir)

    # Find the columns that use enum, boolean, etc. types to create a jOOQ mapping for them
    type_mappings = common.get_type_mappings(args.dbmodel.metadata)

    # Generate the XML config for the jOOQ code generator
    jooq_config_file = code_generator.generate_jooq_config(
        type_mappings, args.dbsettings, schema_name, db_query_package
    )

    if args.do_jooq:
        # Create a temporary database, run the jOOQ generator on it, then remove it
        with args.dbmodel.metadata.db_instance(args.dbsettings, not args.keep_db):
            utils.print_action(
                f"Running jOOQ generator on temporary database {args.dbsettings.name}"
            )

            error = _run_jooq(args.java, jooq_config_file, args.classpath)
            utils.print_end_action(error)
            if error is not None:
                raise error



#!/usr/bin/env python3
# encoding: utf-8
"""
Convenience methods to define store procedures in model files
"""

from . import ValueHolder
from textwrap import dedent, indent


class ProcParamDirection(ValueHolder):
    """
    Hold the direction of a parameter in a stored procedure
    """

    def __init__(self, value):
        assert value in ("IN", "OUT", "INOUT")
        super().__init__(value)


# The available parameter directions:
IN = ProcParamDirection("IN")
OUT = ProcParamDirection("OUT")
INOUT = ProcParamDirection("INOUT")


class ProcSqlBody(ValueHolder):
    """
    Hold the SQL code that makes up the body of a stored procedure
    """

    pass


class ProcComment(ValueHolder):
    """
    Hold the comment test of a store procedure
    """

    pass


class ProcParam(object):
    """
    A parameter in a stored procedure
    """

    def __init__(self, name, the_type, direction):
        self._name = name
        self._type = the_type
        self._direction = None

        self.direction = direction

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        """
        A custom setter to make sure the value is an instance of
        ProcParamDirection
        """
        assert isinstance(value, ProcParamDirection)
        self._direction = value

    @property
    def sql(self):
        """
        The representation of the parameter in SQL
        """
        return "{0.direction.value} {0._name} {0._type}".format(self)


class StoredProcedure(object):
    """
    Models a SQL store procedure
    """

    def __init__(self, name, *args):
        self.name = name
        self.comment = None
        self.params = []
        self.sqlbody = None
        self.process_arguments(args)

    def process_arguments(self, args):
        """
        Arguments to the constructor can come in any order and their semantics
        is determined through their type, following the way a Table in
        sqlalchemy is created.
        """

        for arg in args:
            if isinstance(arg, ProcComment):
                self.comment = arg()
            elif isinstance(arg, ProcSqlBody):
                self.sqlbody = arg()
            elif isinstance(arg, ProcParam):
                self.params.append(arg)
            else:
                raise ValueError(
                    "Unexpected argument type for StoredProcedure: {}".type(arg)
                )

    @property
    def sql(self):
        """
        MySQL statement of procedure creation based on the current procedure
        definition
        """
        indented = lambda x: indent(x, 4 * " ")
        statement = [f"CREATE OR REPLACE PROCEDURE {self.name}("]
        if self.params:
            params_text = ",\n".join((param.sql for param in self.params))
            statement += ["\n", indented(params_text), "\n"]
        statement.append(")\nLANGUAGE SQL")
        if self.comment:
            statement.append(f"\nCOMMENT '{self.comment}'")
        indented_sql = indented(dedent(self.sqlbody))
        statement.append(f"\nBEGIN\n{indented_sql}\nEND\n")
        return "".join(statement)

    @property
    def creation_statement(self):
        return "\n".join(["DELIMITER //", self.sql + "\n//", "DELIMITER ;"])


def test():
    """
    Simple test for definition and text output
    """
    proc = StoredProcedure(
        "topiccounter",
        ProcParam("count", "BIGINT", OUT),
        ProcParam("count2", "BIGINT", OUT),
        ProcComment("Count the topics"),
        ProcSqlBody(
            """
          SELECT COUNT(*) INTO count FROM topic;
        """
        ),
    )
    print(proc.creation_statement)


if __name__ == "__main__":
    test()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ---------------------------------------------------------------------------
# Standard imports:
import os
from datetime import datetime, timezone

# Third party imports
from functools import lru_cache

import pystache

# Local imports
from . import utils
from . import version


# ---------------------------------------------------------------------------


def get_abs_path(relpath):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), relpath))


class NumericEnumTemplateData:
    """
    Hold all necessary data to render all Enum related templates
    """

    CONVERTER_SUFFIX = "TypeConverter"

    def __init__(self, name):
        self.name = name
        self._num = 0
        self._values = []
        self._package = None
        self._conv_package = None
        self._datetime_cache = None

    def compiler_version(self):
        return version.VERSION

    @lru_cache()
    def datetime(self):
        return datetime.now(timezone.utc).isoformat()

    def code_file_name(self):
        return f"{self.name}.java"

    def converter_class_name(self):
        return f"{self.name}{self.CONVERTER_SUFFIX}"

    def converter_file_name(self):
        return f"{self.name}{self.CONVERTER_SUFFIX}.java"

    def enum_fqn(self):
        return f"{self._package}.{self.name}"

    def converter_fqn(self):
        return f"{self._conv_package}.{self.name}{self.CONVERTER_SUFFIX}"

    def _add_value(self, name):
        """
        Add a value by its name and assign an increasing numeric value to it
        """
        self._values.append(dict(valname=name, valnum=self._num))
        self._num += 1
        return self

    def with_values(self, names):
        for name in names:
            self._add_value(name)
        return self

    def with_enum_package(self, package):
        self._package = package
        return self

    def with_enum_converter_package(self, package):
        self._conv_package = package
        return self

    def values(self):
        """
        Return all values as dictionaries of name and numeric value. Mark the last one
        with the 'last' key to help template rendering
        """
        return self._values[:-1] + [dict(self._values[-1], last=True)]

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


class EnumCodeGenerator:
    """
    Code generation logic for Enums and Enum converters
    """

    ENUM_CODE_TEMPLATE = "java_enum"
    ENUM_CONV_CODE_TEMPLATE = "java_enum_converter"
    JOOQ_GEN_CONFIG_TEMPLATE = "jooq_generator_config"
    TEMPLATE_DIR = get_abs_path("./templates")

    def __init__(
        self, enums, code_dir, config_dir, enum_package, converter_package, model_file
    ):
        self.renderer = pystache.Renderer(search_dirs=self.TEMPLATE_DIR)
        self.code_dir = os.path.abspath(code_dir)
        self.config_dir = os.path.abspath(config_dir)
        self.enum_package = enum_package
        self.enum_converter_package = converter_package
        self.database_model_file_name = model_file

        self.enums = self._adjust_enums_from_model(enums)
        self.enums_by_name = {enum.name: enum for enum in self.enums}

    def _adjust_enums_from_model(self, enums):
        """
        Decorate the bare enum definitions from the model with additional attributes
        necessary for code generation
        """
        return [
            NumericEnumTemplateData(enum.name)
            .with_values(enum._values)
            .with_enum_package(self.enum_package)
            .with_enum_converter_package(self.enum_converter_package)
            for enum in enums.values()
        ]

    def _prepare_package_dir(self, package):
        elements = [self.code_dir] + package.split(".")
        directory = os.path.join(*elements)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        return directory

    def _new_java_file(self, package, file_name):
        directory = self._prepare_package_dir(package)
        if not file_name.endswith(".java"):
            file_name += ".java"
        path = os.path.join(directory, file_name)
        return open(path, "w")

    def _new_config_file(self, name):
        if not os.path.isdir(self.config_dir):
            os.makedirs(self.config_dir)
        path = os.path.join(self.config_dir, name)
        return open(path, "w")

    def _render_enum_code(self, enum, package, file_name, template):
        data = {
            "package": package,
            "database_model_file": self.database_model_file_name.replace("\\", "/"),
            "file_name": file_name,
        }
        with self._new_java_file(package, file_name) as jout:
            path = jout.name
            jout.write(self.renderer.render_name(template, enum, data))

        return path

    def render_enum(self, enum):
        return self._render_enum_code(
            enum=enum,
            package=self.enum_package,
            file_name=enum.code_file_name(),
            template=self.ENUM_CODE_TEMPLATE,
        )

    def render_enum_converter(self, enum):
        return self._render_enum_code(
            enum=enum,
            package=self.enum_converter_package,
            file_name=enum.converter_file_name(),
            template=self.ENUM_CONV_CODE_TEMPLATE,
        )

    def render_jooq_gen_config(self, template_data):
        return self.renderer.render_name(self.JOOQ_GEN_CONFIG_TEMPLATE, template_data)

    def _jooq_forced_type_data(self, table, field, type_name):
        return {"expression": f"{table}\\.{field}", "name": type_name}

    def generate_enum_java_code(self):
        utils.print_section_header("Java Enums And Converters")
        for enum in self.enums:
            utils.print_action(f"Generating files for enum: {enum.name}")
            utils.print_generated_file(self.render_enum(enum))
            utils.print_generated_file(self.render_enum_converter(enum))
            utils.print_end_action()

    def generate_jooq_config(self, type_mappings, dbsettings, schema_name, package):
        field_data = []
        for table, columns in type_mappings.items():
            for column, type_name in columns.items():
                type_fqn = type_name
                # If the type is an Enum, fully qualify the name
                enum = self.enums_by_name.get(type_name, None)
                if enum is not None:
                    type_fqn = enum.enum_fqn()
                field_data.append(self._jooq_forced_type_data(table, column, type_fqn))

        template_data = {
            "dbhost": dbsettings.host,
            "dbname": dbsettings.name,
            "dbuser": dbsettings.user,
            "dbpassword": dbsettings.password,
            "dbport": dbsettings.port,
            "schema_name": schema_name,
            "codedir": self.code_dir,
            "fields": field_data,
            "enums": self.enums,
            "package": package,
        }

        utils.print_section_header("jOOQ")
        utils.print_action("Generating jOOQ generator configuration file")
        config_file_name = "ogma_jooq_gen_config.{}.xml".format(schema_name.lower())
        with self._new_config_file(config_file_name) as xout:
            xout.write(self.render_jooq_gen_config(template_data))

        utils.print_generated_file(xout.name)
        utils.print_end_action()

        return xout.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Handling of the enum-tables subcommand
"""
# ---------------------------------------------------------------------------
# Standard imports:
import collections
import re
from functools import lru_cache

# Third party imports
from sqlalchemy import MetaData, Table, Column, Integer, String

# Local imports
from .. import utils
from . import common

# ---------------------------------------------------------------------------

EnumTable = collections.namedtuple("EnumTable", "enum table values")
View = collections.namedtuple("View", "name table selects joins")


def _make_view(from_table):
    view_name = f"enumed_{from_table}_view"
    return View(view_name, from_table, [], [])


@lru_cache(maxsize=100)
def _camel_to_snake(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


@lru_cache(maxsize=100)
def _table_name_from_enum(enum_name):
    return f"enum_{_camel_to_snake(enum_name)}"


def _create_enum_tables(engine, enums):
    metadata = MetaData()
    enum_tables = []
    for enum in enums:
        table_name = _table_name_from_enum(enum.name)
        table = Table(
            table_name,
            metadata,
            Column("value", Integer, nullable=False, unique=True),
            Column("name", String(50), nullable=False, unique=True),
        )
        values = [
            dict(value=value, name=name) for value, name in enumerate(enum._values)
        ]
        enum_tables.append(EnumTable(enum, table, values))

    utils.print_action("Generating tables for enums")
    metadata.drop_all(engine)
    metadata.create_all(engine)

    for enum_table in enum_tables:
        utils.li(f"Enum: {enum_table.enum.name} --> Table: {enum_table.table.name}")
    utils.print_end_action()

    return enum_tables


def _populate_enum_tables(engine, enum_tables, verbose=False):
    utils.print_action("Populating enum tables with values")
    with engine.connect() as conn:
        for enum_table in enum_tables:
            num_values = len(enum_table.values)
            utils.li(f"Table {enum_table.table.name}: {num_values} values")
            if verbose:
                for value in enum_table.values:
                    utils.li(f"    - {value['value']:2}: {value['name']}")
            conn.execute(enum_table.table.insert(), enum_table.values)
    utils.print_end_action()


def _generate_views(engine, enum_tables, enum_usage):
    enum_table_lookup = {et.enum.name: et.table.name for et in enum_tables}
    utils.print_action("Creating views with enum names")
    with engine.connect() as conn:
        for table, fields in enum_usage.items():
            view = _make_view(table)
            for column, enum_name in fields.items():
                enum_table = enum_table_lookup.get(enum_name)
                view.selects.append(f"{enum_table}.name as {column}_name")
                view.joins.append(
                    f"LEFT JOIN {enum_table} ON t.{column} = {enum_table}.value"
                )
            query = "CREATE OR REPLACE VIEW {name} AS SELECT t.*, {selects} FROM {table} t {joins}".format(
                name=view.name,
                table=view.table,
                selects=", ".join(view.selects),
                joins=" ".join(view.joins),
            )

            conn.execute(query)
            utils.li(
                f"View {view.name} to combine table {view.table} with its enum values"
            )
    utils.print_end_action()


def enum_tables(args):

    enums = args.dbmodel.metadata.enums
    if not enums:
        print("No enums found in the given model")
        return

    engine = common.get_db_engine(args.dbsettings)

    # Step 1: Create enum tables
    enum_tables = _create_enum_tables(engine, enums.values())

    # Step 2: Populate enum tables
    _populate_enum_tables(engine, enum_tables, args.verbose)

    # Step 3: Generate views
    eusage = common.get_type_mappings(args.dbmodel.metadata, [common.TypeFamilies.enum])
    _generate_views(engine, enum_tables, eusage)



#!/usr/bin/env python3
# encoding: utf-8

from contextlib import contextmanager as _contextmanager
import copy as _copy
import collections as _collections
from typing import NamedTuple as _NamedTuple

import sqlalchemy as _sqlalchemy
import sqlalchemy.dialects.mysql as _mysql
from sqlalchemy import event as _event


def multiline_rstrip(s: str) -> str:
    """
    Takes a multiline string and returns a copy with trailing spaces from all lines
    removed.
    """
    return "\n".join((l.rstrip() for l in s.splitlines()))


class DatabaseModelException(Exception):
    """
    The base exception for all problems with the model file
    """

    pass


class IntegerEnum(_sqlalchemy.types.TypeDecorator):
    """
    A SqlAlchemy column type that encodes an enum values as an integer
    """

    impl = _sqlalchemy.Integer

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.field_data = None

    def get_constraint(self, column_name):
        return self.field_data.get_constraint(column_name)


class Column(_sqlalchemy.Column):
    """
    Wrapper around sqlalchemy Column to replace 'default' argument, which is
    more succinct and readable, with server_default, which is the one that will
    determine the DDL output.

    The value of the default parameter is also wrapped in a `literal` when
    necessary, so it does not appear quoted in the DDL
    """

    def __init__(self, *args, **kwargs):
        if "server_default" in kwargs:
            raise DatabaseModelException(
                "server_default should not be used directly in columns. Use default instead"
            )
        default = kwargs.pop("default", None)
        if default is not None:
            kwargs["server_default"] = self._wrap_default(default)

        if "autoincrement" not in kwargs:
            kwargs["autoincrement"] = False

        super().__init__(*args, **kwargs)

    @staticmethod
    def _wrap_default(default):
        if type(default) in (bool, int):
            default = _sqlalchemy.literal(default)
        return default


class Table(_sqlalchemy.Table):

    """
    Wrapper around sqlalchemy Table to add object-like access to the column
    names to use in foreign keys

    Once a table is defined, the name of a column in the form
    `table_name.column_name` can be obtained from the model as:
        tables.table_name.column_name
    """

    def __new__(cls, *args, **kwargs):
        return _sqlalchemy.Table.__new__(
            Table, *args, **Table.add_default_settings(kwargs)
        )

    def __init__(self, name, metadata, *args, **kwargs):
        super().__init__(name, metadata, *args, **kwargs)

        # Create an ad-hoc object to represent the table name and its column
        # names
        columns = type("ColumnNames", (object,), {"_table_name": name})
        metadata._table_names.add_table(columns)

        for column in self.columns.values():
            # Add all column names as attributes to the name container
            setattr(columns, column.name, f"{self.name}.{column.name}")

            # Add a constraint for enums (so the integer value is one of the
            # defined enum values)
            if type(column.type) is IntegerEnum:
                self.append_constraint(column.type.get_constraint(column))

    @staticmethod
    def add_default_settings(kwargs):
        # Make sure we have a stable order of settings across runs:
        kwargs = _collections.OrderedDict(**kwargs)

        def set_default(kw, names, default):
            if not set(names).intersection(kw):
                kw[names[0]] = default

        set_default(kwargs, ("mysql_engine",), "InnoDB")
        set_default(
            kwargs,
            (
                "mysql_character_set",
                "mysql_charset",
                "mysql_default_character_set",
                "mysql_default_charset",
            ),
            "utf8mb4",
        )
        set_default(
            kwargs,
            ("mysql_collate", "mysql_default_collate", "mysql_collation"),
            "utf8mb4_general_ci",
        )
        set_default(kwargs, ("mysql_row_format",), "DYNAMIC")

        return kwargs


class DbSettings(_NamedTuple):
    host: str
    name: str
    port: int
    user: str
    password: str


class Enum:
    """
    Database enum (interger backed). Used indirectly in the model files to define
    enumerations that can later be used as table column types.
    """

    def __init__(self, name, metadata, *args, **kwargs):
        """
        Add all values as attributes for object-like access
        """
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self._values = []
        metadata.add_enum(self)

        for index, element in enumerate(args, start=0):
            if hasattr(self, element):
                raise DatabaseModelException(
                    f"{element} enum value already defined for enum {self.name}"
                )
            setattr(self, element, index)
            self._values.append(element)

    def __call__(self):
        """
        Use the sqlalchemy-like syntax of calling the types, and return the
        underlying Enum at this point.
        """
        column_type = IntegerEnum()
        column_type.field_data = self
        return column_type

    def get_constraint(self, column):
        values = ",".join([str(v) for v in range(len(self._values))])
        check_text = f"{column.name} in ({values})"
        check_name = f"ck_{column.table.name}_{column.name}"
        return _sqlalchemy.CheckConstraint(check_text, name=check_name)

    def __repr__(self):
        return f"""Enum('{self.name}', {", ".join(self._values)})"""


def _visitor_mysql_datetime(column_name, column):
    """
    A column visitor to change the type of all DateTimes to MySQL specific DATETIME
    with 3 as parameter
    :param column_name:  The name of the visited column
    :param column: The column object being visited
    """
    if type(column.type) == _sqlalchemy.DateTime:
        column.type = _mysql.DATETIME(fsp=3)


def _adapt_model_to_engine(model, engine, for_text=False):
    """
    Modify the model according to the required changes for each database engine
    :param model: The model module
    :param engine: The name of the target database engine
    :param for_text: Whether this adaptation is happening to generate a text
                     representation of the model (DDL)
    :return: The adapted model
    """
    visitors_per_engine = {"mysql": [_visitor_mysql_datetime]}
    visitors = visitors_per_engine.get(engine)
    if not visitors:
        return model

    model_copy = _copy.deepcopy(model)
    for v in visitors:
        model_copy.visit_columns(v)

    for proc in model_copy.stored_procedures:
        # Emit the delimiter change only when output is text
        ddl = proc.creation_statement if for_text else proc.sql

        # Add the DDL in the queue to run after all the schema has been created
        _event.listen(model_copy, "after_create", _sqlalchemy.DDL(ddl))

    return model_copy


class _EnumCollection(_collections.OrderedDict):
    """
    Provide dictionary and object semantics for an enum collection
    When an enum is added, it will be accessible as:
    >>> enums = _EnumCollection()
    >>> Enum("MyEnumName", enums, "value1", "value2")
    Enum('MyEnumName', value1, value2)
    >>> enums.MyEnumName is enums["MyEnumName"]
    True

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_enum(self, enum):
        if hasattr(self, enum.name):
            raise DatabaseModelException(f"{enum.name} enum already defined")
        setattr(self, enum.name, enum)
        self[enum.name] = enum


class _TableCollection:
    """
    An object that holds, as attributes, each table in the model
    """

    def add_table(self, table):
        if hasattr(self, table._table_name):
            raise DatabaseModelException(f"{table._table_name} table already defined")
        setattr(self, table._table_name, table)


class MetaData(_sqlalchemy.MetaData):
    """
    An augmentation of SqlAlchemy's default MetaaData class to provide some additional
    attributes such as:
     - A collection of all defined tables so far
     - A collection of all defined enums so far
    """

    def __init__(self, *args, **kwargs):
        self.enums = _EnumCollection()

        self._table_names = _TableCollection()
        """A collection with the sole purpose of storing a dummy object for
        each table that contains object-like access to column names"""

        self.stored_procedures = []

        super().__init__(*args, **kwargs)

    def __getstate__(self):
        """
        Control what gets copied with copy.deepcopy.

        __getstate__ must return a dictionary with what will be copied from the
        current instance.

        This is necessary because the parent defines one and it does not include the
        additional attributes of this class.
        """
        parent = super().__getstate__()
        parent.update(
            {
                "enums": self.enums,
                "_table_names": self._table_names,
                "stored_procedures": self.stored_procedures,
            }
        )
        return parent

    def __setstate__(self, state):
        """
        Control what gets copied with copy.deepcopy.

        __setstate__ takes a dictionary (like the one returned from __getstate__)
        with what will be copied from the original instance.

        This is necessary because the parent defines one and it does not include the
        additional attributes of this class.
        """
        super().__setstate__(state)
        self.enums = state["enums"]
        self._table_names = state["_table_names"]
        self.stored_procedures = state["stored_procedures"]

    def add_enum(self, enum):
        self.enums.add_enum(enum)

    def add_stored_procedure(self, stored_procedure):
        self.stored_procedures.append(stored_procedure)

    def visit_columns(self, *visitors):
        """
        Call all given column visitors with every column of every defined table

        The interface of the visitors is:
        def visitor(column_name: str, column: Column) -> None
        """
        for table in self.sorted_tables:
            for column_name, column in table.columns.items():
                for visit in visitors:
                    visit(column_name, column)

    def ddl(self, engine_name="mysql"):
        """
        Generate the DDL text for the currently defined model with the given
        database engine
        """

        class DdlDumper(object):
            def __init__(self):
                self._ddl = []

            def __call__(self, sql, *multiparams, **params):
                self._ddl.append(str(sql.compile(dialect=engine.dialect)).strip())

            @property
            def ddl(self):
                code = ";\n\n".join(self._ddl)
                return multiline_rstrip(code)

        ddl_dumper = DdlDumper()
        engine = _sqlalchemy.create_engine(
            f"{engine_name}://", strategy="mock", executor=ddl_dumper
        )
        model = _adapt_model_to_engine(self, engine_name, for_text=True)
        sorted_tables = [model.tables[t.name] for t in model.sorted_tables]
        model.create_all(engine, checkfirst=False, tables=sorted_tables)
        return ddl_dumper.ddl

    @staticmethod
    def _create_database(engine, database):
        engine.execute(f"CREATE DATABASE {database}")
        engine.execute(f"USE {database}")

    @staticmethod
    def _drop_database(engine, database):
        engine.execute(f"DROP DATABASE {database}")

    def save(self, dbsettings):
        """
        Create the current model, attached to this MetaData instance, in the database
        given in dbsettings.
        """
        db = dbsettings
        connect_string = f"mysql+pymysql://{db.user}:{db.password}@{db.host}:{db.port}"
        engine = _sqlalchemy.create_engine(connect_string)
        self._create_database(engine, dbsettings.name)
        model = _adapt_model_to_engine(self, "mysql")
        model.create_all(engine)
        return engine

    @_contextmanager
    def db_instance(self, dbsettings, drop_after=True):
        """
        A context manager that starts with a created database schema according to the
        model file and drops the whole database after it is exited.
        :param dbsettings: A DbSettings object with the parameters to connect to a
                           database
        :param drop_after: A bool (defaults to True) that triggers the deletion of the
                           database after exiting the context

        The intended usage is:
        >>> with metadata.db_instance(dbsettings):
        >>>     do_operations_with_schema_in_db(dbsettings)
        >>> # The database is deleted here
        """
        engine = self.save(dbsettings)
        try:
            yield
        finally:
            if drop_after:
                self._drop_database(engine, dbsettings.name)



#!/usr/bin/env python3
# encoding: utf-8


class ValueHolder(object):
    """
    Hold one arbitrary value per instance.
    This exists so that a module level variable can be propagated down through
    imports while keeping value changes.

    Instances are callables that return the held value
    """

    def __init__(self, value=None):
        self.value = value

    def __call__(self):
        return self.value



#!/usr/bin/env python3
# encoding: utf-8

from ogma.utils import error
import re
from textwrap import dedent


def _validate_schema_exists(model):
    name = model.schema_name()
    if name is None:
        error(
            dedent(
                f"""\
            Schema name is required in DB model files but could not be found in:
                {model.__file__}
            Specify a schema with:
                Schema("name")"""
            )
        )


def _validate_schema_name(model):
    name = model.schema_name()
    invalid_chars = r"""-^<>/'"{}[\]~`"""
    valid_name = re.compile(fr"""^[^{invalid_chars}]+$""")
    chars = "." + invalid_chars
    if not valid_name.match(name) or "." in name:
        error(
            dedent(
                f"""\
            Invalid schema name:
                {name}
            was found in file:
                {model.__file__}
            A valid schema name cannot contain any of: {chars}"""
            )
        )


def validate_model(model):
    for validator in (_validate_schema_exists, _validate_schema_name):
        validator(model)


class DatabaseModelException(Exception):
    pass


def _get_line(model_content: str, lineno: int):
    assert lineno > 0
    lines = model_content.splitlines(keepends=False)
    return lines[lineno - 1]


def validate_model_file(model_file, allow_imports):
    """
    Take a model file, parse and look for instances of imports.
    :param model_file: The path for the model file
    :param allow_imports: Whether imports are allowed or not
    :raise DatabaseModelException if the model file contains imports and they are not
           allowed
    """
    if not allow_imports:
        import ast

        with open(model_file, "r") as fin:
            code = fin.read()
        tree = ast.parse(code, model_file)
        for node in ast.walk(tree):
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                raise DatabaseModelException(
                    f"Invalid model: import statement found on line {node.lineno}:"
                    f"\n{_get_line(code, node.lineno)}"
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ---------------------------------------------------------------------------
# Standard imports:

# Third party imports

# Local imports
from .. import modelutils
from .. import utils
from . import common

# Command implementations
from .generate import generate
from .enumtables import enum_tables
from .enumusage import enum_usage

# ---------------------------------------------------------------------------

__all__ = ["generate", "enum_usage", "enum_tables", "get_db_name"]


def get_db_name(args):
    print(modelutils.get_new_database_name())


def drop_db(args):
    utils.print_action(f"Dropping database: {args.database}")
    error = None
    try:
        args.dbsettings.name = "mysql"
        engine = common.get_db_engine(args.dbsettings)
        engine.execute(f"DROP DATABASE {args.database}")
    except Exception as ex:
        error = str(ex)
    finally:
        utils.print_end_action(error=error)


def create_db(args):
    utils.print_action(f"Creating database: {args.dbsettings.name}")
    error = None
    try:
        args.dbmodel.metadata.save(args.dbsettings)
    except Exception as ex:
        error = str(ex)
    finally:
        utils.print_end_action(error=error)



#!/usr/bin/env python3
# encoding: utf-8

import uuid as _uuid
from datetime import datetime as _datetime

# DDL elements
from sqlalchemy import (
    # Column,
    ForeignKey,
    ForeignKeyConstraint,
    CheckConstraint,
    UniqueConstraint,
    PrimaryKeyConstraint,
    Index,
    literal,
    text,
)

# Data types for columns
from sqlalchemy import (
    Boolean,
    DateTime,
    Integer,
    BigInteger,
    Numeric,
    String,
    Text,
    LargeBinary,
    VARBINARY,
    BINARY,
)

from .sqlalchemywrappers import (
    IntegerEnum,
    Column,
    Table as _Table,
    Enum as _Enum,
    MetaData,
    DbSettings,
)

from .validation import validate_model

from .value_holder import ValueHolder

from .stored_procedures import (
    StoredProcedure as _StoredProcedure,
    ProcComment,
    ProcSqlBody,
    ProcParam,
    IN,
    OUT,
    INOUT,
)

# Come predefined literals to avoid the use of literal strings for keywords in the
# model files.
CURRENT_TIMESTAMP = text("CURRENT_TIMESTAMP")
NULL = text("NULL")
CASCADE = "CASCADE"
SET_NULL = "SET NULL"
RESTRICT = "RESTRICT"


def get_new_database_name():
    return "_".join(
        ("ogma_db_", _datetime.utcnow().strftime("%Y%m%d%H%M%S"), _uuid.uuid4().hex)
    )


# Each run deals with a single model file for a single schema, so store at this level
# some model information that will be invisible to the model file, but available for
# processing.

# This holds the schema name for the processed model file, which is set by the Schema
# function below:
schema_name = ValueHolder(None)

# The default metadata object for all the Table objects created in the model file.
# SqlAlchemy needs this to be passed around, so this is stored here and used in the
# DDL object creators below to pass to the corresponding SqlAlchemy objects while
# not showing explicitly in the model file.
metadata = MetaData()

# Easy access to the previously defined tables, so one can create indexes or foreign
# keys in the model file without using the table names as strings.
enums = metadata.enums

# Easy programmatic access to the previously defined enums, so they can be referenced
# as column types when creating tables.
tables = metadata._table_names


def Schema(a_schema_name):
    """
    Declare (call only once, at the top) the schema name for the current model file.
    """
    schema_name.value = a_schema_name


def Table(*args, **kwargs):
    """
    Convenience wrapper for SqlAlchemy Table, which transparently passes the default
    metadata object for the current model file.
    """
    name, rest = args[0], args[1:]
    return _Table(name, metadata, *rest, **kwargs)


def Enum(*args, **kwargs):
    """
    A convenience wrapper for Ogma's numeric DB Enum, which transparently passes the
    default metadata object for the current model file.
    """
    name, rest = args[0], args[1:]
    return _Enum(name, metadata, *rest, **kwargs)


def StoredProcedure(*args, **kwargs):
    """
    A convenience wrapper for Ogma's StoredProcedure, which also adds the new
    StoredProcedure to the metadata object for the current model file.
    """
    sp = _StoredProcedure(*args, **kwargs)
    metadata.add_stored_procedure(sp)
    return sp



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['ogma', 'ogma.commands', 'ogma.modelutils', 'ogma.templates', 'ogma.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyMySQL>=0.9,<0.10',
 'SQLAlchemy>=1.3,<1.4',
 'argcomplete>=1.10,<1.11',
 'colorama>=0.4,<0.5',
 'colored>=1.3,<1.4',
 'pystache>=0.5,<0.6']

entry_points = \
{'console_scripts': ['ogma = ogma:main']}

setup_kwargs = {
    'name': 'ogma',
    'version': '0.1.0',
    'description': 'Ogma: A database access code generator for Java',
    'long_description': "# Ogma\n\n> According to legend, he is the inventor of Ogham, the runic language in which Irish Gaelic was first written. \n\n**Ogma** is a database access code generator for Java. It will take your database schema definition written in a Python-based DSL and generate (with jOOQ, among others) the necessary Java code to perform typed queries on that database.\n\nIt can also generate the necessary DDL to create the database structure according to spec.\n\n*Ogma* has been written for MySQL and MariaDB, but could be made to work with other engines that are both supported by SQLAlchemy and jOOQ.\n\n# How to install\n\nJust run the usual for a Python package:\n\n```bash\npip install ogma\n```\n\nThen you can run `ogma`.\n\n# Code generation and other tools\n\n*Ogma* obviously generates code, but it can also do other things. The tool is organized with subcommands.\n\n## Generation\n\nThe `generate` subcommand is used to generate Java code from the model file. The model file is a Python with some restrictions and additions:\n\n1. There is an implicit import of everything from `modelutils`\n1. No imports are allowed\n\n# For Developers\n## Structure\n * `modelutils` contains all the code that is imported into the database model files and that internally deals with model operations.\n * `commands` contains the entry points for the tool's subcommands\n * `templates` contains *mustache* templates for jOOQ's configuration and additional generated Java code\n\n",
    'author': 'Jacobo de Vera',
    'author_email': 'devel@jacobodevera.com',
    'url': 'https://www.github.com/jdevera/ogma',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
